__title__ = "sense2vec"
__version__ = "1.0.0a1"
__summary__ = "Use NLP to go beyond vanilla word2vec"
__uri__ = "https://github.com/explosion/sense2vec"
__author__ = "Explosion AI"
__email__ = "contact@explosion.ai"
__license__ = "MIT"
__release__ = True
